% # we use a single png for all icon types
<meta charset="UTF-8" />

% # we use a single png for all icon types
<link rel="shortcut icon" type="image/png" href="{{ cdn_static }}/v1/datapane-icon-192x192.png"/>
<link rel="shortcut icon" sizes="192x192" href="{{ cdn_static }}/v1/datapane-icon-192x192.png"/>
<link rel="apple-touch-icon" href="{{ cdn_static }}/v1/datapane-icon-192x192.png"/>

<script src="https://kit.fontawesome.com/256c714d9a.js"></script>
<link
  href="https://fonts.googleapis.com/css?family=Inter:400,500,600,700|Inconsolata|Karla|Rubik&display=swap"
  rel="stylesheet"
/>
<link rel="stylesheet" href="{{ cdn_base }}/report/index.css" />
<link rel="stylesheet" href="{{ cdn_base }}/report/tailwind.css" />
<script>
  // analytics config
  // TODO -- pass these in as props for consistency?
  window.dpAppRunner = "{{ app_runner }}" === "True";
  window.dpLocal = true;
  % if events:
    window.dpLocalViewEvent = {{events}};
    window.dpReportId = "{{report_id}}";
    window.dpAuthorId = "{{author_id}}";
  % end

  /* construct the view */
  const htmlHeader = `<style type="text/css">{{ !css_header }}</style>`;
  window.reportProps = { htmlHeader, isLightProse: {{is_light_prose}}, isOrg: false, mode: "VIEW", reportWidthClass: "{{ report_width_class }}" };

  if (!window.dpAppRunner) {
      // pass app data via props if in local (static) mode, otherwise this data is fetched from the server
      window.reportProps["appData"] = { data: { result: {{!app_data}} }};
  }
</script>
<script type="module">
  import { mountReport } from "{{ cdn_base }}/report/index.es.js";
  mountReport(window.reportProps);
</script>
